"""
Module with variable objects

    Classes
    ------------
    Variables: set of Time, Flow and Effort variables
        RotationalVariables: Implementation of Variables in rotational domain
        LinearVariables: Implementation of Variables in translational domain
        ElectricalVariables: Implementation of Variable in rotational domain

"""

from __future__ import annotations
from typing import TYPE_CHECKING
from abc import ABC, abstractmethod
from dataclasses import dataclass
import numpy as np
from scipy import integrate

if TYPE_CHECKING:
    from dozersim.modelling import Model
    from dozersim.path import Path


class Variables(ABC):
    """ Class that contains flow, effort and time variables
        they are generalized variables that are domain independent
        Rotational domain: Effort = Torque and Flow = Speed
        Linear domain: Effort = Force and Flow = Velocity
        Electrical domain: Effort = Voltage and Flow = Current
        Note the only flow and effort are stored in the variable set
        generalized displacement, acceleration and power are calculated

        Attributes
        -----------
            name (str): Name of the variable set
            time (ndarray): Time series
            effort (ndarray): sum of all efforts
            torque (ndarray): alias for effort
            force (ndarray): alias for effort
            voltage (ndarray): alias for effort
            flow (ndarray): sum of all flows
            velocity (ndarray): alias of flow
            current (ndarray): alias of flow
            power (ndarrat): effort*flow1

        Methods
        -----------
            add: add variable to variable set
            get_all_flows: return all flows
            get_all_efforts: return list of efforts

    """
    __slots__ = ('parent', 'name', '_time', 'efforts', 'flows', '_load_case')

    def __init__(self, parent: Path, name: str = "FEV", old_vars: Variables = None) -> None:
        self.name: str = name
        self.parent = parent
        self._time: Time = old_vars._time if old_vars else None
        self.efforts: list[Effort] = []
        self.flows: list[Flow] = []
        self._load_case: str | None = None

    @property
    def time(self):
        return self._time.data

    @property
    def effort(self) -> np.ndarray:
        """ Return total effort variable as ndarray"""
        efforts = np.array([effort.data for effort in self.efforts])
        return efforts.sum(axis=0)

    @property
    def torque(self) -> np.ndarray:
        """ Alias for effort"""
        return self.effort

    @property
    def force(self) -> np.ndarray:
        """ Alias for effort"""
        return self.effort

    @property
    def voltage(self) -> np.ndarray:
        """ Alias for effort"""
        return self.effort

    @property
    def displacement(self) -> np.ndarray:
        """ Return generalized displacement as ndarray"""
        flows = np.array([flow.data[0] for flow in self.flows])
        return flows.sum(axis=0)

    @property
    def flow(self) -> np.ndarray:
        """ Return total flow as ndarray"""
        flows = np.array([flow.data[1] for flow in self.flows])
        return flows.sum(axis=0)

    @property
    def velocity(self) -> np.ndarray:
        """ Alias for flow """
        return self.flow

    @property
    def current(self) -> np.ndarray:
        """ Alias for flow """
        return self.flow

    @property
    def acceleration(self) -> np.ndarray:
        """ Return generalized acceleration as ndarray"""
        flows = np.array([flow.data[2] for flow in self.flows])
        return flows.sum(axis=0)

    @property
    def power(self) -> np.ndarray:
        """ Return power as ndarray"""
        return self.flow * self.effort

    @property
    def energy(self) -> np.ndarray:
        """ Return power as ndarray"""
        return integrate.cumulative_trapezoid(self.power, self.time, initial=0)

    @property
    def add(self) -> dict[str, callable]:
        """ Returns a dictionary with function to add

            Keys
            ------------
            'effort':
            'torque':
            'force':
            'displacement':
            'charge':
            'flow':
            'velocity':
            'current':
            'acceleration':
            'current_rate':
            'time':

        """
        return {'effort': self._add_effort,
                'torque': self._add_effort,
                'force': self._add_effort,
                'voltage': self._add_effort,
                'displacement': self._add_state0,
                'charge': self._add_state0,
                'flow': self._add_state1,
                'velocity': self._add_state1,
                'current': self._add_state1,
                'acceleration': self._add_state2,
                'current_rate': self._add_state2,
                'time': self._add_time
                }

    def _add_effort(self, values: np.ndarray, parent: Model, name: str):
        """ Add variable to container
                Parameters
                ------------
                    variable (Variable): variables that need to be added to variables set
        """
        effort = Effort(name, parent, values)
        self.efforts.append(effort)

    def _add_state0(self, values: np.ndarray, parent: Model, name: str):
        """ Add variable to container
            Parameters
            ------------
            values
            parent
            name

        """
        data = {0: None, 1: None, 2: None}
        data[0] = values
        data[1] = np.gradient(values, self.time.data)
        data[2] = np.gradient(data[1], self.time.data)

        flow = Flow(name, parent, data)
        self.flows.append(flow)

    def _add_state1(self, values: np.ndarray, parent: Model, name: str, init0: float = 0):
        """ Add variable to container
            Parameters
            ------------
            values
            parent
            name
            order

        """
        data = {0: None, 1: None, 2: None}
        data[0] = integrate.cumulative_trapezoid(values, self.time.data, initial=init0)
        data[1] = values
        data[2] = np.gradient(values, self.time.data)

        flow = Flow(name, parent, data)
        self.flows.append(flow)

    def _add_state2(self, values: np.ndarray, parent: Model, name: str, init1: float = 0, init0: float = 0):
        """ Add variable to container
            Parameters
            ------------
            values
            parent
            name
            init0
            init1

        """
        data = {0: None, 1: None, 2: None}
        data[2] = values
        data[1] = integrate.cumulative_trapezoid(values, self.time.data, initial=init1)
        data[0] = integrate.cumulative_trapezoid(data[1], self.time.data, initial=init0)

        flow = Flow(name, parent, data)
        self.flows.append(flow)

    def _add_time(self, values: np.ndarray, model: Model, name: str):
        """ Add variable to container
            Parameters
            ------------
            variable (Variable): variables that need to be added to variables set

        """
        self._time = Time(name, model, values)

    @property
    @abstractmethod
    def state0_label(self) -> str:
        """ Returns flow units"""
        pass

    @property
    @abstractmethod
    def state1_label(self) -> str:
        """ Returns flow units"""
        pass

    @property
    @abstractmethod
    def state2_label(self) -> str:
        """ Returns flow units"""
        pass

    @property
    @abstractmethod
    def effort_label(self) -> str:
        """ Returns effort units"""
        pass

    @property
    def load_case(self) -> str:
        return self._load_case

    @load_case.setter
    def load_case(self, val: str):
        self._load_case = val

    def __repr__(self) -> str:
        return f"Variables({self.name})"


class RotationalVariables(Variables):
    """ Variables in the rotational domain
    implements specific properties of linear variables like label and units"""
    __slots__ = ()

    def __init__(self, parent: Path, name: str = "Rotational FEV", old_vars: Variables = None) -> None:
        super().__init__(parent, name, old_vars)

    @property
    def state1_label(self) -> str:
        return 'Speed in rad/s'

    @property
    def state0_label(self) -> str:
        return 'Angle in rad'

    @property
    def state2_label(self) -> str:
        return 'Acceleration in rad'

    @property
    def effort_label(self) -> str:
        return 'Torque in Nm'

    def __repr__(self) -> str:
        return f"RotationalVariables({self.name})"


class TranslationalVariables(Variables):
    """ Variables in the linear domain
    implements specific properties of linear variables like label and units"""
    __slots__ = ()

    def __init__(self, parent: Path, name: str = "Linear FEV", old_vars: Variables = None) -> None:
        super().__init__(parent, name, old_vars)

    @property
    def state1_label(self) -> str:
        return 'Speed in m/s'

    @property
    def state0_label(self) -> str:
        return 'Distance in m'

    @property
    def state2_label(self) -> str:
        return 'Acceleration in m/s^2'

    @property
    def effort_label(self) -> str:
        return 'Force in N'

    def __repr__(self) -> str:
        return f"LinearVariables({self.name})"


class ElectricalVariables(Variables):
    """ Variables in the electrical domain
    implements specific properties of linear variables like label and units
    """
    __slots__ = ()

    def __init__(self, parent: Path, name: str = "Electrical FEV", old_vars: Variables = None) -> None:
        super().__init__(parent, name, old_vars)

    @property
    def state1_label(self) -> str:
        return 'Current in A'

    @property
    def state0_label(self) -> str:
        return 'Charge in J'

    @property
    def state2_label(self) -> str:
        return 'Rate of current in A/s'

    @property
    def effort_label(self) -> str:
        return 'Voltage in V'

    def __repr__(self) -> str:
        return f"ElectricalVariables({self.name})"


@dataclass(slots=True)
class Variable(ABC):
    name: str
    parent: Model


@dataclass(slots=True)
class Flow(Variable):
    data: dict[int, np.ndarray]


@dataclass(slots=True)
class Effort(Variable):
    data: np.ndarray


@dataclass(slots=True)
class Time(Variable):
    data: np.ndarray



""" This module contains the basic building blocks for creating models

submodules
-----------
paths :
    module that provide the Path class that track the flow of energy through the system.
settings : 
    module that contains the Settings class that is used to built settings objects.
tree : 
    module that contains a Composite design pattern (models and Composite class) that is used to built tree structures.
_variables :
    module that yields the Variable class that contains the flow and effort of the system.

"""
from .settings import Settings



""" Module this contains simulation objects

"""
from .simulation import Simulation, Step
from .constraints import CustomConstraint, PathConstraint
from .objectives import ConstraintObjective, SettingsObjective, PathObjective
from .parameters import SettingsParameter, DatabaseLink




from __future__ import annotations

from dataclasses import dataclass, field

import numpy as np

from dozersim.path.path import Path
from dozersim.settings.settings import Settings
from dozersim.modelling.model import Model


@dataclass(slots=True)
class TrapezoidSettings(Settings):
    Target: str = None
    PeakValue: float = None
    TimeAtPeak: float = None
    TimeToPeak: float = None
    TimeStart: float = None


def evaluate_smooth_trapezoid(paths: tuple[Path], trapezoid: TrapezoidSettings, model: Model):
    path, = paths

    PeakValue = trapezoid.PeakValue
    TimeAtPeak = trapezoid.TimeAtPeak
    TimeToPeak = trapezoid.TimeToPeak
    TimeStart = trapezoid.TimeStart

    time = path.variables.time

    SampleTime = time[1] - time[0]

    if time[-1] < (TimeStart + TimeAtPeak + TimeToPeak * 2):
        raise Exception("Length of the Trapezoid is bigger than the length of time!!")

    tCurve = np.arange(0, TimeToPeak / 2, SampleTime)
    tFlat = np.arange(0, TimeAtPeak, SampleTime)
    Acceleration = 2 * (PeakValue) / TimeToPeak

    n0 = np.zeros(time[time < TimeStart].shape)
    n01a = Acceleration / TimeToPeak * tCurve ** 2
    n01b = PeakValue / 2 + Acceleration * tCurve - Acceleration / TimeToPeak * tCurve ** 2
    n12 = np.full(tFlat.shape, PeakValue)
    n23a = PeakValue - Acceleration / TimeToPeak * tCurve ** 2
    n23b = PeakValue / 2 - Acceleration * tCurve + Acceleration / TimeToPeak * tCurve ** 2

    profile = np.concatenate((n0, n01a, n01b, n12, n23a, n23b))
    n3 = np.zeros(time.size - profile.size)
    profile = np.concatenate((profile, n3))

    path.variables.add[trapezoid.Target](profile, model, "TrapezoidSettings")


def evaluate_trapezoid(paths: tuple[Path], trapezoid: TrapezoidSettings, model: Model):
    path, = paths

    PeakValue = trapezoid.PeakValue
    TimeAtPeak = trapezoid.TimeAtPeak
    TimeToPeak = trapezoid.TimeToPeak
    TimeStart = trapezoid.TimeStart

    time = path.variables.time

    SampleTime = time[1] - time[0]

    if time[-1] < (TimeStart + TimeAtPeak + TimeToPeak * 2):
        raise Exception("Length of the Trapezoid is bigger than the length of time!!")

    tRamp = np.arange(0, TimeToPeak, SampleTime)
    tFlat = np.arange(0, TimeAtPeak, SampleTime)

    n0 = np.zeros(time[time < TimeStart].shape)
    n01 = PeakValue / TimeToPeak * tRamp
    n12 = np.full(tFlat.shape, PeakValue)
    n23 = PeakValue - PeakValue / TimeToPeak * tRamp

    profile = np.concatenate((n0, n01, n12, n23))
    n3 = np.zeros(time.size - profile.size)
    profile = np.concatenate((profile, n3))

    path.variables.add[trapezoid.Target](profile, model, "TrapezoidSettings")


class Trapezoid(Model):
    def __init__(self, name: str = 'constant', settings=TrapezoidSettings(), eval_fun=evaluate_trapezoid):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)



from __future__ import annotations
from dataclasses import dataclass, field
import numpy as np
from numpy import pi
from sympy import Matrix, sin, cos, symbols, lambdify, diag
from dozersim.path.path import Path
from dozersim.settings.settings import Settings
from dozersim.modelling.model import Model
from dozersim.components.materials import Material


@dataclass(slots=True)
class FlywheelSettings(Settings):
    Pressure: float = None
    Temperature: float = None
    AxialGap: float = None
    RadialGap: float = None
    Geometry: GeometryBase = None
    Material: Material = None
    GasConstant: float = 8.315  # [J/mol K] Universal gas constant
    MolecularDiameter: float = 3.68e-10  # [m] @20degC Effective molecular diameter
    BoltzConstant: float = 1.38e-23  # Boltzman constant
    AirMass: float = 4.782e-26  # [kg] Mass Air
    MolarMass: float = 0.0288  # [kg/mol] Air Molar mass
    EOMs: tuple = None

    def __post_init__(self):
        self.EOMs = get_eoms()


    @property
    def mass(self):
        return self.Geometry.calculate_mass(self.Material)

    @property
    def inertia(self):
        return self.Geometry.calculate_inertia(self.Material)


@dataclass(slots=True)
class GeometryBase(Settings):

    def calculate_inertia(self, material: Material) -> tuple:
        pass

    def calculate_mass(self, material: Material) -> tuple:
        pass


@dataclass(slots=True)
class SolidDiskGeometry(GeometryBase):
    Diameter: float = None
    Width: float = None

    def calculate_inertia(self, material: Material) -> tuple:
        Radius = self.Diameter / 2
        Mass, = self.calculate_mass(material)
        InertiaS = 1 / 2 * Mass * Radius ** 2
        InertiaTG = 1 / 12 * Mass * (3 * Radius ** 2 + self.Width ** 2)
        return InertiaS, InertiaTG

    def calculate_mass(self, material: Material) -> tuple:
        MaterialDensity = material.density
        Radius = self.Diameter / 2
        return MaterialDensity * (pi * self.Width * Radius),


@dataclass(slots=True)
class RimAndWebGeometry(GeometryBase):
    Diameter: float = None
    Width: float = None
    InnerDiameter: float = None
    WebWidth: float = None

    def calculate_inertia(self, material: Material) -> tuple:
        OuterRadius = self.Diameter / 2
        InnerRadius = self.InnerDiameter / 2

        TotalMass, RimMass, WebMass = self.calculate_mass(material)

        RimInertiaS = 1 / 2 * RimMass * (OuterRadius ** 2 + InnerRadius ** 2)
        RimInertiaTG = 1 / 12 * RimMass * (3 * (OuterRadius ** 2 + InnerRadius ** 2) + self.Width ** 2)
        WebInertiaS = 1 / 2 * WebMass * InnerRadius ** 2
        WebInertiaTG = 1 / 12 * WebMass * (3 * OuterRadius ** 2 + self.WebWidth ** 2)

        InertiaTG = WebInertiaTG + RimInertiaTG
        InertiaS = WebInertiaS + RimInertiaS  # wheel intertia w.r.t. the spinning axis

        return InertiaS, InertiaTG

    def calculate_mass(self, material: Material) -> tuple:
        MaterialDensity = material.density

        OuterRadius = self.Diameter / 2
        InnerRadius = self.InnerDiameter / 2

        RimVolume = pi * (OuterRadius ** 2 - InnerRadius ** 2) * self.WebWidth
        RimMass = RimVolume * MaterialDensity

        WebVolume = self.WebWidth * pi * InnerRadius ** 2
        WebMass = WebVolume * MaterialDensity

        TotalMass = WebMass + RimMass
        return TotalMass, RimMass, WebMass


def evaluate_flywheel(paths: tuple[Path, Path, Path], flywheel: FlywheelSettings, model: Model):

    WheelPath, FramePath, GimbalPath = paths

    MomentsSTG, MomentsUVW, VelocitiesSTG = flywheel.EOMs

    WheelVelocity = WheelPath.variables.velocity
    WheelAcceleration = WheelPath.variables.acceleration
    FrameVelocity = FramePath.variables.velocity
    FrameAcceleration = FramePath.variables.acceleration
    GimbalAngle = GimbalPath.variables.displacement
    GimbalVelocity = GimbalPath.variables.velocity
    GimbalAcceleration = GimbalPath.variables.acceleration
    InertiaS, InertiaTG = flywheel.Geometry.calculate_inertia(flywheel.Material)

    velocitySTG = VelocitiesSTG(WheelVelocity, WheelAcceleration, FrameVelocity, FrameAcceleration, GimbalAngle,
                                     GimbalVelocity, GimbalAcceleration, InertiaS, InertiaTG)
    momentsSTG = MomentsSTG(WheelVelocity, WheelAcceleration, FrameVelocity, FrameAcceleration, GimbalAngle,
                                 GimbalVelocity, GimbalAcceleration, InertiaS, InertiaTG)
    momentsUVW = MomentsUVW(WheelVelocity, WheelAcceleration, FrameVelocity, FrameAcceleration, GimbalAngle,
                                 GimbalVelocity, GimbalAcceleration, InertiaS, InertiaTG)

    WheelPath.variables.add['effort'](momentsSTG[0][0], model, 'FlywheelSettings moments')
    FramePath.variables.add['effort'](momentsSTG[1][0], model, 'FlywheelSettings moments')
    GimbalPath.variables.add['effort'](momentsSTG[2][0], model, 'FlywheelSettings moments')

    OuterRadius = flywheel.Geometry.Diameter / 2
    Pressure_Pa = flywheel.Pressure * 1e5  # convert bar to pascal
    Temperature_degK = flywheel.Temperature + 273.15  # convert to kelvin

    OuterSurfaceRadius = flywheel.RadialGap + OuterRadius
    AirDensity = Pressure_Pa / (flywheel.GasConstant * Temperature_degK) * flywheel.MolarMass

    DynamicViscosity = (flywheel.BoltzConstant * flywheel.AirMass * Temperature_degK) ** 0.5 / (
                flywheel.MolecularDiameter ** 2 * pi ** (3 / 2))
    KinematicViscosity = DynamicViscosity / AirDensity
    GapRatio = flywheel.AxialGap / OuterRadius

    RadialMoment = []
    AxialMoment = []

    for velocity in WheelVelocity:
        if velocity == 0:
            Cm_a = 0
            Cm_r = 0
        else:
            ## Reynolds Number
            Re = AirDensity * OuterRadius ** 2 * velocity / DynamicViscosity
            Re_g = AirDensity * velocity * OuterRadius * flywheel.RadialGap / DynamicViscosity

            ## Axial Component
            if Re < 3e5:  # Laminar Flow
                delta = 5.5 * (KinematicViscosity / velocity) ** (0.5)
                #         delta = (nu/omega)**(0.5);
                if flywheel.AxialGap < delta:  # Merged Boundary Layers
                    # Regime 1
                    Cm_a = 2 * pi / (GapRatio * Re)
                #             disp('Regime 1')
                else:  # Separate Boundary Layers
                    # Regime 2
                    Cm_a = 3.7 * GapRatio ** (0.1) * Re ** (-0.5)
            #             disp('Regime 2')
            else:  # Turbulent
                #          delta = 0.526*ro*Re**(-0.2);
                delta = AirDensity ** (3 / 5) * (KinematicViscosity / velocity) ** (1 / 5)
                if flywheel.AxialGap < delta:  # Merged Boundary Layers
                    # Regime 3
                    Cm_a = 2 * 0.040 * GapRatio ** (-1 / 6) * Re ** (
                        -0.25)  # For Regime type 3 (Turbulent flow with merged boundary layers -> Small clearance)
                #             disp('Regime 3')
                else:  # Separate Boundary Layers
                    # Regime 4
                    Cm_a = 0.0102 * GapRatio ** (0.1) * Re ** (-1 / 5)
            #             disp('Regime 4')

            ## Radial Component
            if Re_g < 64:  # Laminar
                Cm_r = 8 * OuterSurfaceRadius ** 2 / Re_g * OuterRadius * (OuterRadius + OuterSurfaceRadius)
            elif (Re_g >= 64) and (Re_g < 500):  # Transitional flow
                Cm_r = 2 * (flywheel.RadialGap / OuterRadius) ** 0.3 * Re_g ** (-0.6)
            #         disp('Laminar')
            elif (Re_g >= 500) and (Re_g < 1e4):  # Transitional flow
                Cm_r = 1.03 * (flywheel.RadialGap / OuterRadius) ** 0.3 * Re_g ** -0.5
            #         disp('Turbulent transition')
            else:
                Cm_r = 0.065 * (flywheel.RadialGap / OuterRadius) ** 0.3 * Re_g ** -0.2
            #         disp('Turbulent')

        RadialMoment.append(0.5 * pi * AirDensity * velocity ** 2 * OuterRadius ** 4 * flywheel.Geometry.Width * Cm_r)
        AxialMoment.append(AirDensity * velocity ** 2 * OuterRadius ** 5 * Cm_a)

    Moment = np.array(RadialMoment) + np.array(AxialMoment)
    Moment = Moment * (WheelVelocity >= 0) - Moment * (WheelVelocity < 0)

    WheelPath.variables.add['effort'](Moment, model, "Air friction")


def get_eoms():
    '''Symbolic equations of motion'''
    InertiaS, InertiaTG = symbols('InertiaS,InertiaTG', real=True)
    WheelAngle, WheelVelocity, WheelAcceleration = symbols('WheelAngle,WheelVelocity,WheelAcceleration', real=True)
    FrameAngle, FrameVelocity, FrameAcceleration = symbols('FrameAngle,FrameVelocity,FrameAcceleration', real=True)
    GimbalAngle, GimbalVelocity, GimbalAcceleration = symbols('GimbalAngle,GimbalVelocity,GimbalAcceleration',
                                                              real=True)

    GeneralizedAngles = Matrix([WheelAngle, FrameAngle, GimbalAngle])
    GeneralizedVelocities = Matrix([WheelVelocity, FrameVelocity, GimbalVelocity])
    GeneralizedAccelerations = Matrix([WheelAcceleration, FrameAcceleration, GimbalAcceleration])

    WheelInertia = diag(InertiaS, InertiaTG, InertiaTG)

    JacobianSTG = Matrix([[Matrix([0, 0, 0]), rotation_z(GimbalAngle) * Matrix([0, 1, 0]), Matrix([0, 0, 1])]])
    JacobianWheel = Matrix([[Matrix([1, 0, 0]), rotation_z(GimbalAngle) * Matrix([0, 1, 0]), Matrix([0, 0, 1])]])

    VelocitiesSTG: Matrix = JacobianSTG * GeneralizedVelocities
    VelocitiesWheel: Matrix = JacobianWheel * GeneralizedVelocities

    JacobianVelocities = VelocitiesWheel.jacobian(GeneralizedAngles)

    AccelerationsWheel = JacobianWheel * GeneralizedAccelerations + JacobianVelocities * GeneralizedVelocities

    MomentsSTG = WheelInertia * AccelerationsWheel + tilde(VelocitiesSTG) * WheelInertia * VelocitiesWheel
    MomentsSTG[0] = InertiaS * WheelAcceleration  # Overwrite wheel-axis moment to remove artifacts
    MomentsUVW = rotation_z(GimbalAngle) * MomentsSTG

    Variables = (
    WheelVelocity, WheelAcceleration, FrameVelocity, FrameAcceleration, GimbalAngle, GimbalVelocity, GimbalAcceleration,
    InertiaS, InertiaTG)
    MomentsSTG = lambdify(Variables, MomentsSTG, 'numpy')
    VelocitiesSTG = lambdify(Variables, VelocitiesSTG, 'numpy')
    MomentsUVW = lambdify(Variables, MomentsUVW, 'numpy')

    return MomentsSTG, MomentsUVW, VelocitiesSTG


def rotation_x(phi):
    """
    RX the rotation matrix for the x-axis
    Parameters
    ----------
    phi

    Returns
    -------
    This function uses the angle phi and returns a 3X3 rotation matrix
    """
    return Matrix([[1, 0, 0], [0, cos(phi), -sin(phi)], [0, sin(phi), cos(phi)]])


def rotation_y(phi):
    """
    RY the rotation matrix for the y-axis
    Parameters
    ----------
    phi

    Returns
    -------
    This function uses the angle phi and returns a 3X3 rotation matrix
    """
    return Matrix([[cos(phi), 0, sin(phi)], [0, 1, 0], [-sin(phi), 0, cos(phi)]])


def rotation_z(phi):
    """
    RZ the rotation matrix for the z-axis
    Parameters
    ----------
    phi

    Returns
    -------
    This function uses the angle phi and returns a 3X3 rotation matrix
    """
    return Matrix([[cos(phi), -sin(phi), 0], [sin(phi), cos(phi), 0], [0, 0, 1]])


def tilde(w):
    w1 = w[0]
    w2 = w[1]
    w3 = w[2]

    return Matrix([[0, -w3, w2], [w3, 0, -w1], [-w2, w1, 0]])


class FlyWheel(Model):
    def __init__(self, name: str = 'flywheel', settings=FlywheelSettings(), eval_fun=evaluate_flywheel):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)



from __future__ import annotations

from dataclasses import dataclass, field

import numpy as np

from dozersim.settings.settings import Settings
from dozersim.modelling.model import Model
from dozersim.path.path import Path

@dataclass(slots=True)
class GaussianSettings(Settings):
    Target: str = None
    PeakValue: float = None
    ProfileWidth: float = None
    PeakCentreTime: float = None


def evaluate_gaussian(paths: tuple[Path], gaussian: GaussianSettings, model: Model):
    path, = paths

    time = path.variables.time

    PeakValue = gaussian.PeakValue
    PeakCentreTime = gaussian.PeakCentreTime
    ProfileWidth = gaussian.ProfileWidth

    profile = PeakValue * np.exp(-((time - PeakCentreTime) / ProfileWidth) ** 2)
    path.variables.add[gaussian.Target](profile, model, "GaussianSettings curve")


class Gaussian(Model):
    def __init__(self, name: str = 'constant', settings=GaussianSettings, eval_fun=evaluate_gaussian):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)



from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np

from dozersim.path import path


class Target(ABC):

    @abstractmethod
    def name(self):
        pass

    @abstractmethod
    def evaluate(self, path: path.Path) -> np.ndarray:
        pass

    @abstractmethod
    def units(self, path: path.Path) -> str:
        pass


class PowerTarget(Target):

    def evaluate(self, path: path.Path):
        return path.variables.power

    def units(self, path: path.Path) -> str:
        return "in W"

    def name(self):
        return 'Power'


class EnergyTarget(Target):

    def evaluate(self, path: path.Path):
        return path.variables.energy

    def units(self, path: path.Path) -> str:
        return " in J"

    def name(self):
        return 'Energy'


class EffortTarget(Target):

    def evaluate(self, path: path.Path):
        return path.variables.effort

    def units(self, path: path.Path) -> str:
        return path.variables.effort_label

    def name(self):
        return 'Effort'


class FlowTarget(Target):

    def evaluate(self, path: path.Path):
        return path.variables.flow

    def units(self, path: path.Path) -> str:
        return path.variables.state1_label

    def name(self):
        return 'Flow'


class Statistic(ABC):

    @abstractmethod
    def evaluate(self, values: np.ndarray) -> float:
        pass

    @abstractmethod
    def name(self):
        pass


class Minium(Statistic):

    def evaluate(self, values: np.ndarray):
        return np.min(values)

    def name(self):
        return 'Minimum'


class Maximum(Statistic):

    def evaluate(self, values: np.ndarray):
        return np.max(values)

    def name(self):
        return 'Maximum'


class AbsoluteMinimum(Statistic):

    def evaluate(self, values: np.ndarray):
        return np.min(np.abs(values))

    def name(self):
        return 'AbsoluteMinimum'


class AbsoluteMaximum(Statistic):

    def evaluate(self, values: np.ndarray):
        return np.max(np.abs(values))

    def name(self):
        return 'AbsoluteMaximum'


class Mean(Statistic):

    def evaluate(self, values: np.ndarray):
        return np.mean(values)

    def name(self):
        return 'Mean'


class AbsoluteMean(Statistic):

    def evaluate(self, values: np.ndarray):
        return np.mean(np.abs(values))

    def name(self):
        return 'AbsoluteMean'


class Sum(Statistic):

    def evaluate(self, values: np.ndarray):
        return np.sum(values)

    def name(self):
        return 'Sum'


class SumOfAbsolutes(Statistic):

    def evaluate(self, values: np.ndarray):
        return np.sum(np.abs(values))

    def name(self):
        return 'abssum'


class Nothing(Statistic):

    def evaluate(self, values: np.ndarray):
        return values

    def name(self):
        return 'none'


target_dict = {'power':  PowerTarget(),
               'flow':   FlowTarget(),
               'effort': EffortTarget(),
               'energy': EnergyTarget()}
statistic_dict = {'None': Nothing(),
                  'absmax': AbsoluteMaximum(),
                  'absmin': AbsoluteMinimum(),
                  'max':    Maximum(),
                  'min':    Minium(),
                  'sum':    Sum(),
                  'abssum': SumOfAbsolutes(),
                  'mean': Mean(),
                  'absmean': AbsoluteMean()}



import sys

from PyQt6.QtWidgets import QApplication, QMainWindow, QTabWidget, QWidget, QSlider, QVBoxLayout, QTableWidget, \
    QTableWidgetItem, QHBoxLayout, QLabel, QLineEdit, QSpinBox, QPushButton
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QIntValidator
from matplotlib.backends.backend_qt import NavigationToolbar2QT as NavigationToolbar
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

from dozersim.results.results import Analysis, Result
from dozersim.simulation import parameters, objectives
from dozersim.visualizations.plotting import get_variable_plot, plot_results_3d, plot_results_scatter, \
    plot_results_2d

app = QApplication(sys.argv)


class Visualization(QMainWindow):
    def __init__(self, parent=None, title: str = "Dozer simulation gui"):
        super().__init__()
        self.setWindowTitle(title)
        self.current_window = -1
        self.tabs_main = QTabWidget()
        self.setCentralWidget(self.tabs_main)
        self.resize(1280, 900)
        self._analyses: list[Analysis] = []
        self._analyses_windows: list[AnalysisWindows] = []
        self.show()

    @property
    def analysis(self):
        return self._analyses[-1]

    def add_analysis(self, analysis: Analysis):
        self._analyses.append(analysis)
        analysis_tab = AnalysisWindows(analysis=analysis)
        self.tabs_main.addTab(analysis_tab, f'analysis {len(self._analyses)}')
        self._analyses_windows.append(analysis_tab)

    def add_plot(self, x_object: parameters.Parameter | objectives.Objective,
                 y_object: parameters.Parameter | objectives.Objective,
                 z_object: parameters.Parameter | objectives.Objective = None,
                 plot_type: str = 'line'):
        self._analyses_windows[-1].add_plot(x_object, y_object, z_object, plot_type)


class AnalysisWindows(QTabWidget):

    def __init__(self, analysis: Analysis, parent=None):
        super().__init__(parent)
        self.additional_plots = []
        self._analysis = analysis
        self.result_tabs = ResultWindow(result=analysis.results[-1])

        #  self.plot_variables()
        var_tab = QWidget()
        self.addTab(var_tab, 'Results')
        layout = QVBoxLayout()
        var_tab.setLayout(layout)

        bar = QWidget()
        label = QLabel()
        label.setText(f'Select set between 1 and {len(self._analysis.results)}:')
        self.pushButton = QPushButton()
        self.pushButton.setText('Confirm selection')
        self.pushButton.setMaximumWidth(150)
        self.pushButton.clicked.connect(self.update_plots)
        self.spinBox = QSpinBox()
        self.spinBox.setMinimum(1)
        self.spinBox.setMaximum(len(self._analysis.results))
        # spinBox.setPrefix('set ')
        # spinBox.setSuffix(f' of {len(self._analysis.results)}')
        self.spinBox.setMaximumWidth(150)
        self.spinBox.setMinimumWidth(100)
        bar_layout = QHBoxLayout()
        bar_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        bar_layout.addWidget(label)
        bar_layout.addWidget(self.spinBox)
        bar_layout.addWidget(self.pushButton)
        bar.setLayout(bar_layout)
        layout.addWidget(bar)
        layout.addWidget(self.result_tabs)
        headers, data = analysis.get_analysis_table()
        self.addTab(MyTable(headers=headers, data=data), 'Sets')

    def update_plots(self):
        idx = self.spinBox.value()
        self.result_tabs.update_results(self._analysis.results[idx - 1])

    def add_plot(self, x_object, y_object, z_object=None,
                 plot_type: str = 'line'):
        res_collection = self._analysis
        z_type = issubclass(type(z_object), (parameters.Parameter | objectives.Objective))
        fig = plot_dict[z_type, plot_type](res_collection, x_object, y_object, z_object)
        self.additional_plots.append(fig)
        self.addTab(MyPlot(fig=fig), f'plot {len(self.additional_plots)}')


class ResultWindow(QTabWidget):

    def __init__(self, result: Result, parent=None):
        super().__init__(parent)
        self.plots: dict[tuple, MyPlot] = {}
        self.tables: dict[str, MyTable] = {}
        for load_case in result.load_cases:
            path_tabs = QTabWidget()
            for path in result.paths:
                if path is not None:
                    variables = result.get_variables(load_case, path)
                    variable_tabs = QTabWidget()
                    for var in variables:
                        plot = MyPlot(fig=get_variable_plot(var))
                        variable_tabs.addTab(plot, var.name)
                        self.plots.update({(load_case, path, var.name): plot})
                    path_tabs.addTab(variable_tabs, path.name)
            headers, data = result.get_result_table(load_case=load_case)
            table = MyTable(headers=headers, data=data)
            path_tabs.addTab(table, 'Results')
            self.tables.update({load_case: table})
            self.addTab(path_tabs, load_case)

    def update_results(self, result: Result):
        for load_case in result.load_cases:
            for path in result.paths:
                if path is not None:
                    variables = result.get_variables(load_case, path)
                    for var in variables:
                        get_variable_plot(var, fig=self.plots[(load_case, path, var.name)].sc.figure)
                        self.plots[(load_case, path, var.name)].sc.figure.canvas.draw_idle()
            headers, data = result.get_result_table(load_case=load_case)
            self.tables[load_case].update_table(headers, data)

plot_dict = {
    (False, 'line'): plot_results_2d,
    (False, 'scatter'): plot_results_scatter,
    (True, 'line'): plot_results_3d
}


class MyPlot(QWidget):

    def __init__(self, parent=None, fig: Figure = None):
        super().__init__()
        self.sc = FigureCanvas(figure=fig)
        layout = QVBoxLayout()
        self.setLayout(layout)
        new_toolbar = NavigationToolbar(self.sc)
        layout.addWidget(self.sc)
        layout.addWidget(new_toolbar)



def show_plots():
    app.exec()


class MyTable(QWidget):
    def __init__(self, headers, data, parent=None):
        super().__init__()
        layout = QVBoxLayout()
        self.setLayout(layout)
        self.table = QTableWidget()
        self.table.setRowCount(len(data))
        self.table.setColumnCount(len(data[0]))
        for i, row in enumerate(data):
            for j, item in enumerate(row):
                q_item = QTableWidgetItem(item)
                self.table.setItem(i, j, q_item)
        # table.resizeColumnsToContents()
        self.table.resizeRowsToContents()
        self.table.setHorizontalHeaderLabels(headers)
        self.table.move(0, 0)
        layout.addWidget(self.table)

    def update_table(self, headers, data):
        self.table.clear()
        self.table.setRowCount(len(data))
        self.table.setColumnCount(len(data[0]))
        for i, row in enumerate(data):
            for j, item in enumerate(row):
                q_item = QTableWidgetItem(item)
                self.table.setItem(i, j, q_item)
        # table.resizeColumnsToContents()
        self.table.resizeRowsToContents()
        self.table.setHorizontalHeaderLabels(headers)
        self.table.move(0, 0)



"""
Module: path_management

This module defines the `Path` class, which models the energy flow of a system
and manages variables and analysis associated with it. A `Path` acts as a
builder that creates variable objects, handles transformations, and contains
travelers of various types.

Classes
-------
Path
    Represents a path object that maps the energy flow of the modeled system.

Attributes
----------
fev_dict : dict
    A mapping of energy domains ('rotational', 'translational', 'electrical')
    to their respective variable classes.

Usage Example
-------------
>>> from dozersim.path import Path
>>> path = Path("rotational", "RotationalPath")
>>> path.transform("electrical", "ElectricalTransform")
>>> variables, analysis = path.unload()
"""

from __future__ import annotations

from typing import List, Any

from dozersim.variables import variables
from dozersim.results import results

fev_dict = {
    'rotational': variables.RotationalVariables,
    'translational': variables.TranslationalVariables,
    'electrical': variables.ElectricalVariables
}


class Path:
    """
    Represents a paths object that maps the energy flow of the modeled system.
    A `Path` can create and transform variable objects and contains travelers
    of various types such as analysis and constraints.

    Attributes
    ----------
    name : str
        Name of the paths.
    verbose : bool
        Indicates whether verbose output is enabled.
    travelers : list
        List of travelers (e.g., loads, constraints).
    _variables : list[variables.Variables]
        List of variables representing the energy flow in the paths.
    _start_type : str
        Initial domain type ('rotational', 'translational', 'electrical').

    Methods
    -------
    reset()
        Clears the paths's variables and travelers.
    get_result_object(*travtype: type | tuple) -> List
        Retrieves a list of travelers of specified types.
    add_result_object(traveler: analysis.Load | analysis.ConstraintValue | analysis.ObjectiveValue)
        Adds a traveler to the paths.
    transform(var_type: str, name: str)
        Transforms the energy flow to a new domain by creating a new variable set.
    unload() -> tuple[list[variables.Variables], list[Any]]
        Unloads and returns all variables and travelers, then resets the paths.
    check(var_type: str)
        Checks if the latest variable set matches the specified domain type.
    variables -> variables.Variables
        Returns the latest variable set.
    """

    def __init__(self, start_type: str, name: str = "Path", static: bool = False) -> None:
        """
        Initializes a new Path instance.

        Parameters
        ----------
        start_type : str
            Initial domain type ('rotational', 'translational', 'electrical').
        name : str, optional
            Name of the paths (default is "Path").
        """
        self.name: str = name
        self.static = static
        self.verbose = False
        self._variables: list[variables.Variables] = []
        self._start_type = start_type
        self.travelers: list = []
        self.reset()

    def reset(self):
        """
        Clears the paths's variables and travelers, and resets it to its initial state.
        """
        self._variables = [fev_dict[self._start_type](self, name='source')]
        self.travelers = []

    def get_result_object(self, *travtype: type | tuple) -> List[
        results.Load | results.ConstraintValue | results.ObjectiveValue]:
        """
        Retrieves a list of travelers of the specified types.

        Parameters
        ----------
        travtype : type or tuple
            Types of travelers to retrieve (e.g., Load, ConstraintValue, ObjectiveValue).

        Returns
        -------
        list
            List of travelers matching the specified types.
        """
        return list(filter(lambda traveler: isinstance(traveler, travtype), self.travelers))

    def add_result_object(self, traveler: results.Load | results.ConstraintValue | results.ObjectiveValue):
        """
        Adds a traveler to the paths.

        Parameters
        ----------
        traveler : Load | ConstraintValue | ObjectiveValue
            The traveler to add.
        """
        self.travelers.append(traveler)

    def transform(self, var_type: str, name: str):
        """
        Transforms the energy flow by creating a new set of variables.
        Acts as a transformer if the domain matches the previous one, or as a
        gyrator if it changes domains.

        Parameters
        ----------
        var_type : str
            New domain type ('rotational', 'translational', 'electrical').
        name : str
            Name of the new variable set.
        """
        variable_type = fev_dict[var_type]
        if self._variables:
            variables = variable_type(parent=self, name=name, old_vars=self._variables[-1])
        else:
            variables = variable_type(parent=self, name=name)
        self._variables.append(variables)

    def unload(self) -> tuple[list[variables.Variables], list[Any]]:
        """
        Unloads and returns all variables and result objects, then resets the paths.

        Returns
        -------
        tuple
            variables : list[variables.Variables]
                List of all flow and effort variables.
            result_objects : list[Any]
                List of all constraints and objective values.
        """
        variables = self._variables
        result_objects = self.get_result_object(results.ObjectiveValue, results.ConstraintValue)
        return variables, result_objects

    def check(self, var_type: str):
        """
        Checks if the latest variable set matches the specified domain type.

        Parameters
        ----------
        var_type : str
            Domain type to check ('rotational', 'translational', 'electrical').

        Raises
        ------
        Exception
            If the domain type does not match.
        """
        if not isinstance(self._variables[-1], fev_dict[var_type]):
            raise Exception("You're in the wrong domain!")

    @property
    def variables(self) -> variables.Variables:
        """
        Returns the latest variable set.

        Returns
        -------
        variables.Variables
            The latest variable set.
        """
        return self._variables[-1]

    def __repr__(self) -> str:
        """
        Returns a string representation of the paths.

        Returns
        -------
        str
            String representation of the paths.
        """
        return f"Path({self.name})"



"""
Module: settings_management

This module defines the `Model` class, which provides a framework for managing
settings and constraints for a simulation. It allows dynamic addition and retrieval
of settings and constraints and defines a customizable evaluation function.

Classes
-------
Model
    Abstract settings class for managing settings and constraints in a simulation.

Attributes
----------
_constraints : dict[str, constraints.Constraint]
    Dictionary of constraints mapped by their names.
eval_fun : callable
    Evaluation function used for custom logic.

Usage Example
-------------
>>> from dozersim.settings import Settings
>>> part = Settings()
>>> part.set_setting("max_speed", 100)
>>> print(part.get_setting("max_speed"))
100
"""
from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field



@dataclass
class Settings(ABC):
    """
    Abstract settings class for managing simulation part and constraints.

    Attributes
    ----------
    _constraints : dict[str, constraints.Constraint]
        Dictionary of constraints, where keys are constraint names and values
        are `Constraint` objects.
    eval_fun : callable
        Function for custom evaluation logic.
    """

    def set_setting(self, keyword: str, value):
        """
        Dynamically sets a setting attribute.

        Parameters
        ----------
        keyword : str
            Name of the setting to set.
        value : Any
            Value to assign to the setting.
        """
        setattr(self, keyword, value)

    def get_setting(self, keyword: str):
        """
        Retrieves the value of a specified setting attribute.

        Parameters
        ----------
        keyword : str
            Name of the setting to retrieve.

        Returns
        -------
        Any
            Value of the specified setting.

        Raises
        ------
        AttributeError
            If the specified setting does not exist.
        """
        try:
            return getattr(self, keyword)
        except AttributeError as e:
            raise AttributeError(e)

    @property
    def supplier_table(self):
        """
        Placeholder for a supplier table. Can be overridden in subclasses.

        Returns
        -------
        None
        """
        return None



from __future__ import annotations
from dataclasses import dataclass, field
from dozersim.path.path import Path
from dozersim.settings.settings import Settings
from dozersim.modelling.model import Model
from dozersim.simulation import constraints
from dozersim.components import tools


@dataclass(slots=True)
class GearboxSettings(Settings):
    name: str = None
    ratio: float = None
    max_torque: float = None
    efficiency: float = None
    mass: float = None


def evaluate_gearbox(paths: tuple[Path], gearbox: GearboxSettings, model: Model):
    path = paths[0]

    path.check('rotational')

    torques_output = path.variables.efforts
    torque_output_total = path.variables.torque
    speed_output = path.variables.velocity

    model.constraints['peak_torque'].evaluate(model=model, path=path)

    path.transform(var_type='rotational', name=f"{model.name} input")

    speed_input = speed_output * gearbox.ratio

    path.variables.add['velocity'](speed_input, model, "GearboxSettings input")

    for TorqueOutput in torques_output:
        path.variables.add['torque'](TorqueOutput.data / gearbox.ratio, model, TorqueOutput.name)

    friction_torque = tools.calculate_friction_torque(effort=path.variables.effort,
                                                      power=path.variables.power,
                                                      efficiency=gearbox.efficiency
                                                      )

    path.variables.add['torque'](friction_torque, model, "GearboxSettings friction")


class Gearbox(Model):
    def __init__(self, name: str = 'gearbox', settings=GearboxSettings(), eval_fun=evaluate_gearbox):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)
        self.add_constraint(constraints.PathConstraint(name='peak_torque', settings=settings, ub='max_torque',
                                                       statistic='absmax', target='effort'))





import matplotlib
import numpy as np
from dozersim.results.results import Analysis
from dozersim.simulation import parameters, objectives
from PyQt6.QtWidgets import QWidget
import matplotlib.pyplot as plt
from matplotlib.figure import Figure


matplotlib.use('QtAgg')


def get_variable_plot(var, fig: Figure = None) -> Figure:
    if fig is not None:
        axs = fig.axes
        for ax in axs:
            ax.cla()
    else:
        fig, axs = plt.subplots(4, sharex='col')  # .add_subplot(111)
    nflows = 0
    n_efforts = 0
    for flow in var.flows:
        axs[0].plot(var.time, flow.data[0], label=flow.name)
        axs[0].grid(True)
        axs[0].set_ylabel(var.state0_label)
        axs[0].set_title(f"FEV: {var.name}")
        axs[1].plot(var.time, flow.data[1], label=flow.name)
        axs[1].grid(True)
        axs[1].set_ylabel(var.state1_label)
        # axs[0].set_title(f"FEV: {var.name}")
        nflows += 1
    if nflows > 1:
        axs[1].plot(var.time, var.flow, label='Total')
    axs[1].legend()
    for effort in var.efforts:
        axs[2].plot(var.time, effort.data, label=effort.name)
        axs[2].grid(True)
        axs[2].set_ylabel(var.effort_label)
        n_efforts += 1
    if n_efforts > 1:
        axs[2].plot(var.time, var.effort, label='Total')
    axs[2].legend()
    if (nflows > 0) & (n_efforts > 0):
        axs[3].plot(var.time, var.power)
        axs[3].grid(True)
        axs[3].set_ylabel('Power in W')
        axs[3].set_xlabel('Time in s')

    return fig


def plot_results_2d(analysis: Analysis,
                    x_object: parameters.Parameter | objectives.Objective,
                    y_object: parameters.Parameter | objectives.Objective | list,
                    *args, **kwargs) -> Figure:
    """
    Plot a single

    Parameters
    ----------
    y_object
    x_object

    Returns
    -------

    """
    fig, ax = plt.subplots()  # .add_subplot(111)

    for load_case in analysis.load_cases:
        x_data = np.array(analysis.get_result_values(load_case, x_object))
        y_data = np.array(analysis.get_result_values(load_case, y_object))
        '''
        if type(y_object) is not list:
            y_object = [y_object]
        for obj in y_object:
            y_data = np.array(analysis.get_result_series(load_case, obj))
            plt.plot(x_data, y_data, label=load_case.name)
        '''
        ax.plot(x_data, y_data, label=load_case)
        ax.set_ylabel(y_object.name)
        ax.set_xlabel(x_object.name)
        ax.legend()
        plt.suptitle(f"Load case {load_case}")
    return fig

    # for load_case in analysis.load_cases:
    #    headers, data = analysis.get_result_sets(load_case)


def plot_results_3d(analysis: Analysis,
                    x_object: parameters.Parameter | objectives.Objective,
                    y_object: parameters.Parameter | objectives.Objective,
                    z_object: parameters.Parameter | objectives.Objective) -> Figure:
    """
    Plot a single

    Parameters
    ----------
    analysis :
    z_object
    y_object
    x_object

    Returns
    -------

    """
    fig, ax = plt.subplots(1, 1, subplot_kw={'projection': '3d'})
    for load_case in analysis.load_cases:
        x_data = np.array(analysis.get_result_values(load_case, x_object))
        y_data = np.array(analysis.get_result_values(load_case, y_object))
        x_length = len(list(x_object))
        y_length = len(list(y_object))
        x_data = x_data.reshape((x_length, y_length))
        y_data = y_data.reshape((x_length, y_length))
        z_data = np.array(analysis.get_result_values(load_case, z_object))
        z_data = z_data.reshape((x_length, y_length))

        ax.set_xlabel(x_object.name)
        ax.set_ylabel(y_object.name)
        ax.set_zlabel(z_object.name)
        ax.plot_surface(x_data, y_data, z_data, rstride=1, cstride=1, cmap='viridis', edgecolor='none')
        #ax.plot_trisurf(x_data, y_data, z_data, cmap=cm.jet, linewidth=0.1)
        ax.legend()
        plt.draw()

    return fig


def plot_results_scatter(analysis: Analysis, x_object: objectives.Objective,
                         y_object: objectives.Objective,
                         *args, **kwargs) -> Figure:
    """
    Plot a single

    Parameters
    ----------
    analysis
    y_object
    x_object

    Returns
    -------

    """
    fig, ax = plt.subplots()
    legend_texts = []

    for load_case in analysis.load_cases:
        x_data = np.array(analysis.get_result_values(load_case, x_object))
        y_data = np.array(analysis.get_result_values(load_case, y_object))

        ax.scatter(np.round(x_data, 2), np.round(y_data, 2))

        set_label = 1
        for x_val, y_val in zip(x_data, y_data):
            ax.annotate(f'Set {set_label}', (x_val, y_val))
            set_label += 1

        legend_texts.append(load_case)

    ax.set_xlabel(x_object.name)
    ax.set_ylabel(y_object.name)
    ax.legend(legend_texts)
    plt.suptitle(f"Trade off")

    return fig

#        headers, data = analysis.get_result_sets()






from __future__ import annotations
from dataclasses import dataclass, field
from dozersim.path.path import Path
from dozersim.settings.settings import Settings
from dozersim.modelling.model import Model


@dataclass(slots=True)
class InertiaSettings(Settings):
    name: str = None
    inertia: float = None


def evaluate_inertia(paths: tuple[Path], inertia: InertiaSettings, model: Model):
    path = paths[0]

    path.check('rotational')

    acceleration = path.variables.acceleration

    # Calculate and add rotor inertia moment
    moment_inertia = inertia.inertia * acceleration
    path.variables.add['effort'](moment_inertia, model, "Motor InertiaSettings")
    # paths.variable.add_effort(moment_intertia, model, "Generic InertiaSettings")


class Inertia(Model):
    def __init__(self, name: str = 'gearbox', settings=InertiaSettings(), eval_fun=evaluate_inertia):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)




from __future__ import annotations

import json
import os
from dataclasses import dataclass, field

import numpy as np
from scipy import interpolate

from dozersim.settings.settings import Settings
from dozersim.modelling.model import Model
from dozersim.variables.variables import Flow
from dozersim.path.path import Path

__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))


@dataclass(slots=True)
class Sit2StandDataSettings(Settings):
    Data: dict = None
    Joint: str = None
    SubjectBodyMass: float = None

    def __post_init__(self):
        with open(os.path.join(__location__, '../data/sit2stand_data.json')) as file_object:
            self.Data: dict = json.load(file_object)


def evaluate_sit2stand_data(paths: tuple[Path], sit2stand: Sit2StandDataSettings, model: Model):
    path, = paths

    path.check('rotational')
    old_time = sit2stand.Data["Time"]

    target = sit2stand.Joint
    torque_target = "Torque" + target
    angle_target = "Angle" + target
    speed_target = "Speed" + target
    torque_function = interpolate.interp1d(x=old_time, y=sit2stand.Data[torque_target], kind='cubic')
    angle_function = interpolate.interp1d(x=old_time, y=sit2stand.Data[angle_target], kind='cubic')
    speed_function = interpolate.interp1d(x=old_time, y=sit2stand.Data[speed_target], kind='cubic')

    new_time = path.variables.time
    torque = torque_function(new_time)
    angle = angle_function(new_time)
    speed = speed_function(new_time)
    acceleration = np.gradient(speed, new_time)

    data = {
        0: angle * np.pi / 180,
        1: speed * np.pi / 180,
        2: acceleration * np.pi / 180
    }
    flow = Flow(name=target, parent=model, data=data)
    path.variables.flows.append(flow)
    path.variables.add['torque'](torque * sit2stand.SubjectBodyMass, model, torque_target)


class Sit2StandData(Model):

    def __init__(self, name: str = 'sit2standdata', settings=Sit2StandDataSettings(), eval_fun=evaluate_sit2stand_data):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)


from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
import numpy as np
from numpy import pi
from scipy import interpolate
from dozersim.suppliers import suppliers
from dozersim.modelling.model import Model
from dozersim.settings.settings import Settings
from dozersim.path.path import Path
from dozersim.results.results import Load
from dozersim.simulation import constraints



@dataclass(slots=True)
class BallBearingSettings(Settings):
    name: str = None
    built: str = None
    limitingSpeed: float = None
    boreDiameter: float = None
    outerDiameter: float = None
    width: float = None
    mass: float = None
    staticLoadRating: float = None
    dynamicLoadRating: float = None
    calculationFactor: float = None
    minimumLoadFactor: float = None
    preLoad: float = None
    grease: GreaseSettings = None
    desired_life = None
    ReplenishmentConstant = 6 * 10e-8  # replenishment/starvation constant for grease lubrication
    MovementFrictionCoeff = 0.15  # constant depending on movement


    @property
    def mean_diameter(self) -> float:
        return (self.outerDiameter + self.boreDiameter) / 2

    @property
    def supplier_table(self):
        return suppliers.BallBearingTable


@dataclass(slots=True)
class GreaseSettings(Settings):
    name: str = None
    base: str = None
    viscosityAt40Deg: float = None
    viscosityAt100Deg: float = None
    operatingTemperature: float = None

    @property
    def Viscosity(self):
        v40 = self.viscosityAt40Deg
        v100 = self.viscosityAt100Deg
        TemperatureKelvin = self.operatingTemperature + 273
        C1 = (np.log10(np.log10(v100 + 0.7)) - np.log10(np.log10(v40 + 0.7))) / (
                    np.log10(40 + 273) - np.log10(100 + 273))
        C2 = np.log10(np.log10(v100 + 0.7)) + C1 * np.log10(100 + 273)
        return 10 ** (10 ** (C2 - C1 * np.log10(TemperatureKelvin))) - 0.7

    @property
    def supplier_table(self):
        return suppliers.GreaseTable


def evaluate_ball_bearing(paths: list[Path], bearing: BallBearingSettings, model: Model):
    path = paths[0]

    path.check('rotational')
    speed = np.abs(path.variables.flow * 30 / pi)
    time = np.asarray(path.variables.time)

    # Add preload to bearing
    path.add_result_object(Load(parent=model, path=path,
                                force_radial=np.full(time.shape, 0),
                                force_axial=np.full(time.shape, bearing.preLoad)))

    # Extract bearing loads from paths
    loads = path.get_result_object(Load)
    forceRadial = sum(np.array(load.force_radial) for load in loads)
    forceAxial = sum(np.array(load.force_axial) for load in loads)

    grease = bearing.grease
    viscosity = grease.Viscosity
    if 'mineral' in grease.base:
        SlidingFrictionCoeff = 0.05
    elif 'synthetic' in grease.base:
        SlidingFrictionCoeff = 0.04

    calculator = calculator_dict[bearing.built]

    Grr, Gsl, Kz = calculator.GetFrictionVariables(bearing, forceRadial, forceAxial, speed)
    meanDiameter = bearing.mean_diameter
    boreDiameter = bearing.boreDiameter
    outerDiameter = bearing.outerDiameter

    ## Inlet shear heating reduction
    phi_ish = 1 / (1 + 1.84 * 10 ** (-9) * (np.abs(speed) * meanDiameter) ** 1.28 * viscosity ** 0.64)
    phi_ish = np.sign(speed) * phi_ish
    # Kinematic replenishment/starvation reduction factor
    phi_rs = 1 / np.exp(bearing.ReplenishmentConstant * viscosity * np.abs(speed) * (boreDiameter + outerDiameter)
                        * np.sqrt(np.abs(Kz) / (2 * (outerDiameter - boreDiameter))))
    phi_rs = np.sign(speed) * np.sign(Kz) * phi_rs
    MomentRollingFriction = phi_ish * phi_rs * Grr * (viscosity * speed) ** 0.6 * 1e-3

    phi_bl = 1 / (np.exp(2.6 * 10e-8 * (speed * viscosity) ** 1.4 * meanDiameter))

    mu_sl = phi_bl * bearing.MovementFrictionCoeff + (1 - phi_bl) * SlidingFrictionCoeff

    MomentSlidingFriction = Gsl * mu_sl * 10 ** (-3)

    FrictionMoment = (MomentRollingFriction + MomentSlidingFriction) * 2

    FrictionMoment = FrictionMoment * (speed >= 0) - FrictionMoment * (speed < 0)


    # Calculate static and dynamic bearing limits
    MinimumRadialLoad = bearing.minimumLoadFactor * (viscosity * np.mean(np.abs(speed)) / 1000) ** (2 / 3) * (
                meanDiameter / 100) ** 2

    P, P0 = calculator.GetEquivalentLoad(bearing, forceRadial, forceAxial)

    LifeRating10 = (bearing.dynamicLoadRating / P) ** 3
    NumberOfRotations = np.trapz(time, np.abs(speed) / 60)
    RotationTime = time[-1]
    RotationPerHour = NumberOfRotations / RotationTime * 60 * 60
    LifeRating10h = 10 ** 6 / RotationPerHour * LifeRating10

    path.variables.add['effort'](FrictionMoment, model, 'Bearing friction')
    # Add contraints here
    model.constraints['bearing_life'].evaluate(model=model, path=path, value=LifeRating10)
    model.constraints['static_load'].evaluate(model=model, path=path, value=P0)


class BallBearingCalculator(ABC):
    @abstractmethod
    def GetFrictionVariables(self, settings: BallBearingSettings, ForceRadial, ForceAxial, Speed) -> tuple:
        pass

    @abstractmethod
    def GetEquivalentLoad(self, settings: BallBearingSettings, forceRadial, forceAxial) -> tuple:
        pass


class DeepGrooveCalculation(BallBearingCalculator):

    def __init__(self) -> None:
        # Creating table for e, X and Y
        AxialFactorTable = [0.175, 0.345, 0.689, 1.03, 1.38, 2.07, 3.45, 5.17, 6.86]
        LoadRatioLimitTable = [0.19, 0.22, 0.26, 0.28, 0.30, 0.34, 0.38, 0.42, 0.44, ]
        XFactorTable = [0.56, 0.56, 0.56, 0.56, 0.56, 0.56, 0.56, 0.56, 0.56]
        YFactorTable = [2.3, 1.99, 1.71, 1.55, 1.45, 1.31, 1.15, 1.04, 1.00]
        self.LoadRatioFunction = interpolate.interp1d(x=AxialFactorTable, y=LoadRatioLimitTable, kind='cubic',
                                                      fill_value='extrapolate')
        self.XFactorFunction = interpolate.interp1d(x=AxialFactorTable, y=XFactorTable, kind='cubic',
                                                    fill_value='extrapolate')
        self.YFactorFunction = interpolate.interp1d(x=AxialFactorTable, y=YFactorTable, kind='cubic',
                                                    fill_value='extrapolate')

    def GetFrictionVariables(self, settings: BallBearingSettings, ForceRadial, ForceAxial, Speed) -> tuple:
        ## Numbers based on 619 and 639 series bearing
        Kz = 3.1  # Geometric constant

        R1 = 4.3e-7  # Geometric constant
        R2 = 1.7  # Geometric constant

        # Sliding friction variable
        S1 = 4.75e-3  # Geometric constant
        S2 = 3.6  # Geometric constant

        if max(ForceAxial) == 0:
            Grr = R1 * settings.mean_diameter ** 1.97 * ForceRadial ** 0.54
            Gsl = S1 * settings.mean_diameter ** -0.26 * ForceRadial ** (5 / 3)
        else:
            F_a = np.abs(ForceAxial)
            alphaF = (24.6 * (F_a / settings.staticLoadRating) ** 0.24) * pi / 180
            Grr = R1 * settings.mean_diameter ** 1.97 * (ForceRadial + R2 / np.sin(alphaF) * F_a) ** 0.54
            Gsl = S1 * settings.mean_diameter ** -0.145 * (
                    ForceRadial ** 5 + S2 * settings.mean_diameter / np.sin(alphaF) * F_a ** 4) ** (1 / 3)

        return (Grr*np.sign(ForceAxial), Gsl*np.sign(ForceAxial), Kz*np.sign(ForceAxial))

    def GetEquivalentLoad(self, settings: BallBearingSettings, forceRadial, forceAxial) -> tuple:
        Fr = np.mean(np.abs(forceRadial))
        Fa = np.mean(np.abs(forceAxial))
        Fr0 = np.max(np.abs(forceRadial))
        Fa0 = np.max(np.abs(forceAxial))

        LoadRatio = Fa / Fr
        AxialFactor = settings.calculationFactor * Fa / settings.staticLoadRating

        if Fr < 0.1:  ## In case of 'pure' axial loads
            if settings.boreDiameter <= 12:
                X0 = 0
                Y0 = 4
            else:
                X0 = 0
                Y0 = 2
        else:
            X0 = 0.5
            Y0 = 0.6

        RatioLimit = self.LoadRatioFunction(AxialFactor)
        if LoadRatio <= RatioLimit:
            EquivalentDynamicLoad = Fr
        else:
            X = self.XFactorFunction(AxialFactor)
            Y = self.YFactorFunction(AxialFactor)
            EquivalentDynamicLoad = X * Fr + Y * Fa

        EquivalentStaticLoad = X0 * Fr0 + Y0 * Fa0
        if EquivalentStaticLoad < Fr0:
            EquivalentStaticLoad = Fr0

        return (EquivalentDynamicLoad, EquivalentStaticLoad)


class AngularContactCalculation(BallBearingCalculator):

    def GetFrictionVariables(self, settings: BallBearingSettings, ForceRadial, ForceAxial, Speed) -> tuple:
        Kz = 4.4  # Geometric constant

        # Rolling friction variable
        R1 = 5.03e-7  # Geometric constant
        R2 = 1.97  # Geometric constant
        R3 = 1.90e-12  # Geometric constant
        Fg = R3 * settings.mean_diameter ** 4 * Speed ** 2

        Grr = R1 * settings.mean_diameter ** 1.97 * (ForceRadial + Fg + R2 * ForceAxial) ** 0.54

        # Sliding friction variable
        S1 = 1.30e-2  # Geometric constant
        S2 = 0.68;  # Geometric constant
        S3 = 1.91e-12  # Geometric constant
        Fg = S3 * settings.mean_diameter ** 4 * Speed ** 2;  #
        Gsl = S1 * settings.mean_diameter ** 0.26 * ((ForceRadial + Fg) ** (4 / 3) + S2 * ForceAxial ** (4 / 3))

        return (Grr, Gsl, Kz)

    def GetEquivalentLoad(self, settings: BallBearingSettings, forceRadial, forceAxial) -> tuple:
        Fr = np.mean(np.abs(forceRadial))
        Fa = np.mean(np.abs(forceAxial))
        Fr0 = np.max(np.abs(forceRadial))
        Fa0 = np.max(np.abs(forceAxial))

        if 'B' in settings.name:  # For 40 deg contact angle (single)
            loadRatioLimit = 1.14
            factorX = 0.35
            factorY2 = 0.57
            factorY0 = 0.26
        elif 'AC' in settings.name:  # for 25 deg contact angle (single)
            loadRatioLimit = 0.68
            factorX = 0.41
            factorY2 = 0.87
            factorY0 = 0.38
        else:
            raise Exception('Bearing name should contain contact angle specification (B or AC)')

        LoadRatio = Fa / Fr
        if LoadRatio <= loadRatioLimit:
            EquivalentDynamicLoad = Fr
        else:
            X = factorX
            Y2 = factorY2
            EquivalentDynamicLoad = X * Fr + Y2 * Fa

        X0 = 0.5
        Y0 = factorY0
        EquivalentStaticLoad = X0 * Fr0 + Y0 * Fa0
        if EquivalentStaticLoad < Fr0:
            EquivalentStaticLoad = Fr0

        return (EquivalentDynamicLoad, EquivalentStaticLoad)


class SuperPrecisionCalculation(BallBearingCalculator):

    def __init__(self) -> None:
        # For contact angle 15 deg
        AxialFactorTable = [0.178, 0.357, 0.714, 1.07, 1.43, 2.14, 3.57, 5.35]
        LoadRatioLimitTable = [0.38, 0.4, 0.43, 0.46, 0.47, 0.50, 0.55, 0.56]
        X2FactorTable = [0.44, 0.44, 0.44, 0.44, 0.44, 0.44, 0.44, 0.44]
        Y2FactorTable = [1.47, 1.40, 1.3, 1.23, 1.19, 1.12, 1.02, 1.00]
        Y0FactorTable = [0.46, 0.46, 0.46, 0.46, 0.46, 0.46, 0.46, 0.46]
        self.LoadRatioFunction = interpolate.interp1d(x=AxialFactorTable, y=LoadRatioLimitTable, kind='cubic')
        self.X2FactorFunction = interpolate.interp1d(x=AxialFactorTable, y=X2FactorTable, kind='cubic')
        self.Y2FactorFunction = interpolate.interp1d(x=AxialFactorTable, y=Y2FactorTable, kind='cubic')
        self.Y0FactorFunction = interpolate.interp1d(x=AxialFactorTable, y=Y0FactorTable, kind='cubic')
        self.FrictionVariableFunction = AngularContactCalculation.GetFrictionVariables

    def GetFrictionVariables(self, settings: BallBearingSettings, ForceRadial, ForceAxial, Speed) -> tuple:
        return self.FrictionVariableFunction(self, settings, ForceRadial, ForceAxial, Speed)

    def GetEquivalentLoad(self, settings: BallBearingSettings, forceRadial, forceAxial) -> tuple:
        Fr = np.mean(np.abs(forceRadial))
        Fa = np.mean(np.abs(forceAxial))
        Fr0 = np.max(np.abs(forceRadial))
        Fa0 = np.max(np.abs(forceAxial))

        LoadRatio = Fa / Fr
        AxialFactor = settings.calculationFactor * Fa / settings.staticLoadRating
        X0 = 0.5
        if 'AC' in settings.name:
            # Contact angle 25 deg
            Y0 = 0.38
            X2 = 0.41
            Y2 = 0.87
            RatioLimit = 0.68
        else:
            # Contact angle 15 deg
            Y0 = self.Y0FactorFunction(AxialFactor)
            X2 = self.X2FactorFunction(AxialFactor)
            Y2 = self.Y2FactorFunction(AxialFactor)
            RatioLimit = self.LoadRatioFunction(AxialFactor)

        if LoadRatio <= RatioLimit:
            EquivalentDynamicLoad = Fr
        else:
            EquivalentDynamicLoad = X2 * Fr + Y2 * Fa

        EquivalentStaticLoad = X0 * Fr0 + Y0 * Fa0
        if EquivalentStaticLoad < Fr0:
            EquivalentStaticLoad = Fr0

        return (EquivalentDynamicLoad, EquivalentStaticLoad)


calculator_dict = {
        'deep groove': DeepGrooveCalculation(),
        'angular contact': AngularContactCalculation(),
        'super precision': SuperPrecisionCalculation()
        }


class BallBearing(Model):
    def __init__(self, name: str = 'ballbearing', settings=BallBearingSettings(), eval_fun=evaluate_ball_bearing):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)
        self.add_constraint(constraints.CustomConstraint(name='static_load', settings=settings, ub='staticLoadRating'))
        self.add_constraint(constraints.CustomConstraint(name='bearing_life', settings=settings, lb='desired_life'))


from __future__ import annotations

from dataclasses import dataclass, field

import numpy as np

from dozersim.path.path import Path
from dozersim.settings.settings import Settings
from dozersim.modelling.model import Model


@dataclass(slots=True)
class ConstantSettings(Settings):
    Target: str = None
    Magnitude: float = None


def evaluate_constant(paths: tuple[Path], constant: ConstantSettings, model: Model):
    path, = paths

    time = path.variables.time
    profile = np.full(time.shape, constant.Magnitude)
    path.variables.add[constant.Target](profile, model, f"{model.name}")


class Constant(Model):
    def __init__(self, name: str = 'constant', settings=ConstantSettings(), eval_fun=evaluate_constant):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)


"""
This module defines the `Model` class, an abstract base class representing
a simulation model with configurable settings, evaluation functions, and constraints.

Classes
-------
Model
    Abstract base class for creating simulation models.
"""


from dozersim.settings import Settings
from dozersim.path import Path
from dozersim.simulation.constraints import Constraint
from abc import ABC, abstractmethod
from collections import UserList


class Model(ABC):
    """Abstract base class representing a simulation model.

    Attributes
    ----------
    name : str
        Name of the model.
    _settings : Settings
        Simulation settings object.
    _eval_fun : callable
        Evaluation function to run simulations.
    post_eval : list[callable]
        List of post-evaluation functions.
    _constraints : dict[str, Constraint]
        Dictionary storing constraints by name.
    """

    def __init__(self, eval_fun: callable, name: str, settings: Settings) -> None:
        """Initializes the Model with evaluation function, name, and settings.

        Parameters
        ----------
        eval_fun : callable
            Function used to evaluate paths.
        name : str
            Name of the model.
        settings : Settings
            Simulation settings object.
        """
        self.name = name
        self._settings = settings
        self._eval_fun = eval_fun
        self.post_eval: list[callable] = []
        self._constraints: dict[str, Constraint] = {}

    def __repr__(self) -> str:
        """Returns a string representation of the Model.

        Returns
        -------
        str
            A representation string showing the model's name and settings.
        """
        return f'settings.Lead("{self.name}", function, settings)'

    def __str__(self) -> str:
        """Returns the model's name as a string.

        Returns
        -------
        str
            The model's name.
        """
        return self.name

    @property
    def settings(self):
        """Gets the current simulation settings.

        Returns
        -------
        Settings
            The current simulation settings object.
        """
        return self._settings

    @settings.setter
    def settings(self, settings):
        """Sets new simulation settings.

        Parameters
        ----------
        settings : Settings
            New simulation settings object.

        Raises
        ------
        Exception
            If the provided settings are not an instance of `Settings`.
        """
        if isinstance(settings, Settings):
            self._settings = settings
        else:
            raise Exception("Settings should be of Settings class")

    def evaluate(self, paths: list[Path]):
        """Evaluates paths using the evaluation function and settings.

        Parameters
        ----------
        paths : list[Path]
            List of paths to evaluate.
        """
        self._eval_fun(paths, self._settings, self)
        for fun in self.post_eval:
            fun(paths)

    def add_constraint(self, constraint: Constraint):
        """Adds a constraint to the internal constraint dictionary.

        Parameters
        ----------
        constraint : constraints.Constraint
            The constraint to add. Its name is used as the dictionary key.
        """
        cons_name = constraint.name
        self._constraints[cons_name] = constraint

    @property
    def constraints(self):
        """Retrieves all constraints in the internal dictionary.

        Returns
        -------
        dict[str, constraints.Constraint]
            Dictionary mapping constraint names to `Constraint` objects.
        """
        return self._constraints



from __future__ import annotations
from dataclasses import dataclass, field
import numpy as np

from dozersim.path.path import Path
from dozersim.settings.settings import Settings
from dozersim.modelling.model import Model
from dozersim.simulation import constraints


@dataclass(slots=True)
class ServoDriveSettings(Settings):
    name: str = None
    ContinuousCurrent: float = None
    PeakCurrent: float = None
    MaxVoltage: float = None
    SwitchingFrequency: float = None
    SupplyVoltage: float = None
    BusUtilizationFactor: float = np.sqrt(3) / 2


def evaluate_servo_drive(paths: tuple[Path], servo: ServoDriveSettings, model: Model):
    path = paths[0]

    path.check('electrical')

    motor_power = path.variables.power

    supply_voltage = np.full(np.array(motor_power).shape, servo.SupplyVoltage)
    supply_current = motor_power / supply_voltage
    utilization_loss = servo.SupplyVoltage*servo.BusUtilizationFactor

    model.constraints['rated_current'].evaluate(model=model, path=path)
    model.constraints['peak_current'].evaluate(model=model, path=path)
    model.constraints['max_voltage'].evaluate(model=model, path=path)

    path.transform(var_type='electrical', name="Supply variables")
    path.variables.add['effort'](supply_voltage, model, 'Supply Voltage')
    path.variables.add['current'](supply_current, model, 'Supply Current')


class ServoDrive(Model):

    def __init__(self, name: str = 'servodrive', settings=ServoDriveSettings(), eval_fun=evaluate_servo_drive):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)
        self.add_constraint(constraints.PathConstraint(name='rated_current', settings=settings, ub='ContinuousCurrent',
                                                       statistic='absmean', target='flow'))
        self.add_constraint(constraints.PathConstraint(name='peak_current', settings=settings, ub='PeakCurrent',
                                                       statistic='absmax', target='flow'))
        self.add_constraint(constraints.PathConstraint(name='max_voltage', settings=settings, ub='MaxVoltage',
                                                       statistic='absmax', target='effort'))




from __future__ import annotations
from dataclasses import dataclass, field
import numpy as np
from dozersim.path.path import Path
from dozersim.settings.settings import Settings
from dozersim.modelling.model import Model
from dozersim.results.results import Load


@dataclass(slots=True)
class SuspensionSettings(Settings):
    suspension_width: float = None
    target_path: Path = None


def evaluate_suspension(paths: tuple[Path], suspension: SuspensionSettings, model: Model):
    """
        The suspension items simulates a bearing suspension
        that converts a pure moment of the perpendicular paths to a force load on the parallel paths
        It assumes the first paths carries the parallel axis and the remaining two the perpendicular


    """

    moments = []
    if suspension.target_path in paths:
        for path in paths:
            path.check('rotational')
            if path is not suspension.target_path:
                moments.append(path.variables.effort)
    else:
        raise Exception('The _target paths specified in the settings is not found in the paths attached to the items')

    moment = np.linalg.norm(moments, axis=0)

    radial_force = moment / suspension.suspension_width
    axial_force = np.full(np.asarray(moment).shape, 0)

    suspension.target_path.add_result_object(Load(parent=model, path=suspension.target_path,
                                                  force_radial=radial_force, force_axial=axial_force))


class Suspension(Model):
    def __init__(self, name: str = 'suspension', settings=SuspensionSettings(), eval_fun=evaluate_suspension):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)



from __future__ import annotations

from dataclasses import dataclass, field

import numpy as np

from dozersim.path.path import Path
from dozersim.settings.settings import Settings
from dozersim.modelling.model import Model


@dataclass(slots=True)
class TimeSourceSettings(Settings):
    TimeStart: float = None
    TimeEnd: float = None
    SampleTime: float = None


def evaluate_time_source(paths: tuple[Path], setting: TimeSourceSettings, model: Model):
    for path in paths:
        path.variables.add['time'](np.arange(setting.TimeStart, setting.TimeEnd, setting.SampleTime), model, 'Time')


class TimeSource(Model):

    def __init__(self, name: str = 'time', settings: TimeSourceSettings = TimeSourceSettings(), eval_fun=evaluate_time_source):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)



import numpy as np


def calculate_friction_torque(effort: np.ndarray, power: np.ndarray, efficiency: float) -> np.ndarray:
    index = 0
    friction_torque = []
    for pow in power:
        if pow >= 0:  # if forward driving
            friction_torque.append(effort[index] * (1 - efficiency))
        elif pow < 0:  # If backdriving
            friction_torque.append(effort[index] * (1 - 1 / efficiency))
        index += 1
    return np.array(friction_torque)
